import typing as tp
import math

Number = tp.Union[float, int]
Predicate = tp.Callable[[Number], bool]


def between(left: Number = -math.inf, right: Number = math.inf,
            incl_left: bool = True, incl_right: bool = True) -> Predicate:
    """
    Build a predicate to check whether a given number is in particular range

    :param left: predicate will be true for numbers larger than this
    :param right: predicate will be true for numbers smaller than this
    :param incl_left: whether to include left in the range for the predicate. Set to True
        will result in a <= operator, whereas False will result in a >
    :param incl_right: whether to include left in the range for the predicate
    :return: a callable(Number)->bool
    """
    def predicate(x: Number) -> bool:
        if incl_left:
            if x < left:
                return False
        else:
            if x <= left:
                return False

        if incl_right:
            if x > right:
                return False
        else:
            if x >= right:
                return False

        return True
    return predicate



import itertools
import typing as tp

from .decorators import wraps

T = tp.TypeVar('T')
U = tp.TypeVar('U')


# noinspection PyPep8Naming
def _NOP(x):
    return x


def execute_before(callable_: tp.Callable) -> tp.Callable:
    """
    Wrapper to create wrappers which execute callable before function launch.

    Use like this:

    >>> @execute_before
    >>> def do_things():
    >>>     print('Things are done')

    Then the following will print 'Things are done'
    >>> @do_things
    >>> def nothing():
    >>>     ...
    >>> nothing()

    You can even specify custom parameters for the callable:

    >>> @execute_before
    >>> def i_am_2(two):
    >>>     assert two == 2
    >>> @i_am_2(2)
    >>> def run_me():
    >>>     pass
    """

    def outer(*args, **kwargs):
        if len(args) == 1 and not kwargs and callable(args[0]):
            @wraps(args[0])
            def inner(*my_args, **my_kwargs):
                callable_()
                return args[0](*my_args, **my_kwargs)

            return inner
        else:
            def inner(func):
                @wraps(func)
                def inner2(*my_args, **my_kwargs):
                    callable_(*args, **kwargs)
                    return func(*my_args, **my_kwargs)
                return inner2
            return inner

    return outer


def auto_adapt_to_methods(decorator):
    """
    Allows you to use the same decorator on methods and functions,
    hiding the self argument from the decorator.

    Usage:

    >>> @auto_adapt_to_methods
    >>> def times_two(fun):
    >>>     def outer(a):
    >>>         return fun(a*2)
    >>>     return outer
    >>> class Test:
    >>>     @times_two
    >>>     def twice(self, a):
    >>>         return a*2
    >>> @times_two
    >>> def twice(a):
    >>>     return a*2
    >>> assert Test().twice(2) == 4
    >>> assert twice(2) == 4
    """

    def adapt(func):
        return _MethodDecoratorAdaptor(decorator, func)

    return adapt


class _MethodDecoratorAdaptor:
    __slots__ = ('decorator', 'func')

    def __init__(self, decorator, func):
        self.decorator = decorator
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.decorator(self.func)(*args, **kwargs)

    def __get__(self, instance, owner):
        return self.decorator(self.func.__get__(instance, owner))


def attach_arguments(*args, **kwargs):
    """
    Return a decorator that passes extra arguments to the function.

    Example:

    >>> @attach_arguments(2, label='value')
    >>> def print_args(*args, **kwargs):
    >>>     print(args, kwargs)
    >>> print_args(3, 4, key='value')

    will print

    >>> (3, 4, 2) {'key': 'value', 'label': 'value'}

    Arguments given in attach_arguments will take precedence in case of key collisions.
    """

    def outer(fun):
        @wraps(fun)
        def inner(*my_args, **my_kwargs):
            my_kwargs.update(kwargs)
            return fun(*my_args, *args, **my_kwargs)

        return inner

    return outer


def for_argument(*t_ops: tp.Callable[[T], U], **t_kwops: tp.Callable[[T], U]):
    """
    Calls a callable for each of the arguments. Pass None if you do not wish to process given
    argument.

    returns is a special keyword, a callable to process the result through

    Use like:

    >>> @for_argument(int, str, typed=bool, returns=int)
    >>> def check(val1, val2, typed='True'):
    >>>     if typed:
    >>>         return val1 + int(val2)
    """
    t_ops = [_NOP if op == 'self' else op for op in t_ops]
    returns = t_kwops.pop('returns', _NOP)

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method
            assert len(args) >= len(t_ops)
            a = fun(*((_NOP if op is None else op)(arg) for arg, op in
                      itertools.zip_longest(args, t_ops, fillvalue=None)),
                    **{k: t_kwops.get(k, _NOP)(v) for k, v in kwargs.items()})
            return returns(a)

        return inner

    return outer



from .proxy import Proxy


class HashableWrapper(Proxy):
    """
    A class that makes given objects hashable by their id, if the object is already not hashable.

    Note that this class will return a proxy to the object, and not the object itself.

    Use like:

    >>> a = {1:2, 3:4}
    >>> a = HashableWrapper(a)
    >>> hash(a)
    """
    __slots__ = ()

    def __hash__(self):
        my_object = getattr(self, '_Proxy__obj')
        try:
            return hash(my_object)
        except TypeError:
            return hash(id(my_object))



import typing as tp

T = tp.TypeVar('T')
IteratorOrIterable = tp.Union[tp.Iterator[T], tp.Iterable[T]]


def choose_one(filter_fun: tp.Callable[[T], bool], iterable: IteratorOrIterable) -> T:
    """
    Syntactic sugar for

    >>> choose(filter_fun, iterable, True)

    :param filter_fun: function that returns bool on the single value
    :param iterable: iterable to examine
    :return: single element in the iterable that matches given input
    :raises ValueError: on multiple elements matching, or none at all
    """
    return choose(filter_fun, iterable, True)


def choose(filter_fun: tp.Callable[[T], bool], iterable: IteratorOrIterable,
           check_multiple: bool = False) -> T:
    """
    Return a single value that exists in given iterable.

    Essentially the same as:

    >>> next(iter(filter(filter_fun, iterable)))

    but raises a different exception if nothing matches (and if there are multiple matches
    and check_multiple is True).
    If check_multiple is True this guarantees to exhaust the generator (if passed).

    :param filter_fun: function that returns bool on the single value
    :param iterable: iterable to examine
    :param check_multiple: if True, this will check if there are multiple entries matching
        filter_fun, and will raise ValueError if so
    :return: single element in the iterable that matches given input
    :raises ValueError: on multiple elements matching (if check_multiple), or none at all
    """
    elem_candidate = None
    found = False
    for elem in iterable:
        if filter_fun(elem):
            if not check_multiple:
                return elem

            if found:
                raise ValueError(
                    'Multiple values (%s, %s) seen' % (repr(elem_candidate), repr(elem)))
            elem_candidate = elem
            found = True

    if not found:
        raise ValueError('No elements matching given filter seen')

    return elem_candidate



import ctypes
import platform
import threading
import time
import typing as tp
from threading import Condition as PythonCondition

from satella.time import measure
from ...exceptions import ResourceLocked, WouldWaitMore


class Condition(PythonCondition):
    """
    A wrapper to faciliate easier usage of Pythons' threading.Condition.

    There's no need to acquire the underlying lock, as wait/notify/notify_all do it for you.

    This happens to sorta not work on PyPy. Use at your own peril. You have been warned.
    """

    def wait(self, timeout: tp.Optional[float] = None):
        """
        Wait for condition to become true.

        :param timeout: timeout to wait. None is default and means infinity
        :raises ResourceLocked: unable to acquire the underlying lock within specified timeout.
        :raises WouldWaitMore: wait's timeout has expired
        """
        with measure() as measurement:
            if timeout is None:
                self.acquire()
            else:
                if not self.acquire(timeout=timeout):
                    raise ResourceLocked('internal lock locked')

            try:
                if timeout is None:
                    super().wait()
                else:
                    if not super().wait(timeout=timeout - measurement()):
                        raise WouldWaitMore('wait was not notified')
            finally:
                self.release()

    def notify_all(self) -> None:
        """
        Notify all threads waiting on this Condition
        """
        with self._lock:
            super().notify_all()

    def notify(self, n: int = 1):
        """
        Notify n threads waiting on this Condition

        :param n: amount of threads to notify
        """
        with self._lock:
            super().notify(n=n)


class SingleStartThread(threading.Thread):
    """
    A thread that keeps track of whether it's .start() method was called, and does nothing
    if it's called second or so time.
    """

    def __init__(self, *args, **kwargs):
        self.__started = False
        super().__init__(*args, **kwargs)

    def start(self) -> 'SingleStartThread':
        """
        No-op when called second or so time. The first time it starts the thread.

        :return: self
        """
        if self.__started:
            return
        self.__started = True
        super().start()
        return self


class TerminableThread(threading.Thread):
    """
    Class that will execute something in a loop unless terminated. Use like:

    >>> class MeGrimlock(TerminableThread):
    >>>     def loop(self):
    >>>         ... do your operations ..
    >>> a = MeGrimlock().start()
    >>> a.terminate().join()

    Flag whether to terminate is stored in **self.terminating**.

    If you decide to override run(), you got to check periodically for **self._terminating**
    to become true. If it's true, then a termination request was received, and the thread should
    terminate itself.
    If you decide to use the loop/cleanup interface, you don't need to do so, because it will
    be automatically checked for you before each loop() call.

    You may also use it as a context manager. Entering the context will start the thread, and
    exiting it will .terminate().join() it, in the following way:

    >>> a = MeGrimlock()
    >>> with a:
    >>>     ...
    >>> self.assertFalse(a.is_alive())
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._terminating = False  # type: bool

    @property
    def terminating(self) -> bool:
        """Return whether a termination of this thread was requested"""
        return self._terminating

    def loop(self) -> None:
        """
        Run one iteration of the loop. Meant to be overrided. You do not need to override it
        if you decide to override run() through.

        This should block for as long as a single check will take, as termination checks take place
        between calls.
        """

    def start(self) -> 'TerminableThread':
        """
        Start the execution of this thread
        :return: this thread
        """
        super().start()
        return self

    def run(self) -> None:
        """
        Calls self.loop() indefinitely, until terminating condition is met
        """
        while not self._terminating:
            self.loop()
        self.cleanup()

    def cleanup(self):
        """
        Called after thread non-forced termination, in the thread's context.

        The default implementation does nothing.
        """

    def __enter__(self):
        """Returns self"""
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate().join()
        return False

    def safe_sleep(self, interval: float, wake_up_each: float = 2) -> bool:
        """
        Sleep for interval, waking up each wake_up_each seconds to check if terminating,
        finish earlier if is terminating.

        :param interval: Time to sleep in total
        :param wake_up_each: Amount of seconds to wake up each
        :return: status of _terminating flag at the exit
        """
        t = 0
        while t < interval and not self._terminating:
            remaining_to_sleep = min(interval - t, wake_up_each)
            time.sleep(remaining_to_sleep)
            t += remaining_to_sleep
        return self._terminating

    def terminate(self, force: bool = False) -> 'TerminableThread':
        """
        Signal this thread to terminate.

        Forcing, if requested, will be done by injecting a SystemExit exception into target
        thread, so the thread must acquire GIL. For example, following would not be interruptable:

        >>> time.sleep(1000000)

        Note that calling force=True on PyPy won't work, and NotImplementedError will be raised
        instead.

        :param force: Whether to force a quit
        :return: self
        :raises RuntimeError: when something goes wrong with the underlying Python machinery
        :raises NotImplementedError: force=True was used on PyPy
        """
        self._terminating = True
        if force:
            if platform.python_implementation() == 'PyPy':
                raise NotImplementedError('force=True was made on PyPy')
            ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident),
                                                             ctypes.py_object(SystemExit))
            if ret == 0:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Zero threads killed!')
            elif ret > 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Multiple threads killed!')

        return self



import typing as tp
import warnings
from queue import Queue


def queue_iterator(queue: Queue) -> tp.Iterator:
    """
    Syntactic sugar for

    >>> while queue.qsize() > 0:
    >>>     yield queue.get()
    """
    while queue.qsize() > 0:
        yield queue.get()


def update_if_not_none(dictionary: dict, key, value):
    """
    Deprecated alias for :func:`update_key_if_none`
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0,'
                  'use update_key_if_not_none instead', DeprecationWarning)
    return update_key_if_none(dictionary, key, value)


def update_attr_if_none(obj: object, attr: str, value: tp.Any,
                        on_attribute_error: bool = True,
                        if_value_is_not_none: bool = False) -> None:
    """
    Updates the object attribute, if it's value is None, or if
    it yields AttributeError (customizable as per on_attribute_error parameter)

    :param obj: object to alter
    :param attr: attribute to set
    :param value: value to set
    :param on_attribute_error: whether to proceed with setting the value on
        AttributeError while trying to read given attribute. If False, AttributeError
        will be raised.
    :param if_value_is_not_none: update object unconditionally, if only value is not None
    """
    if if_value_is_not_none:
        if value is not None:
            setattr(obj, attr, value)
    else:
        try:
            val = getattr(obj, attr)
            if val is None:
                setattr(obj, attr, value)
        except AttributeError:
            if on_attribute_error:
                setattr(obj, attr, value)
            else:
                raise


def update_key_if_none(dictionary: dict, key, value):
    """
    This is deprecated. Please use update_key_if_not_none instead!
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0, use '
                  'update_key_if_not_none instead', DeprecationWarning)
    return update_key_if_not_none(dictionary, key, value)


def update_key_if_not_none(dictionary: dict, key, value):
    """
    Syntactic sugar for

    >>> if value is not None:
    >>>     dictionary[key] = value

    :param dictionary: dictionary to update
    :param key: key to use
    :param value: value to use
    :return: the dictionary itself
    """
    if value is not None:
        dictionary[key] = value
    return dictionary



from .atomic import AtomicNumber
from .callablegroup import CallableGroup, CallNoOftenThan
from .locked_dataset import LockedDataset
from .locked_structure import LockedStructure
from .monitor import MonitorList, Monitor, MonitorDict, RMonitor
from .thread import TerminableThread, Condition, SingleStartThread
from .id_allocator import IDAllocator

__all__ = ['LockedDataset', 'Monitor', 'RMonitor', 'CallableGroup', 'TerminableThread',
           'MonitorDict', 'MonitorList', 'Condition', 'LockedStructure', 'AtomicNumber',
           'CallNoOftenThan', 'SingleStartThread', 'IDAllocator']



from .dump_frames_on import install_dump_frames_on
from .trace_back import Traceback, GenerationPolicy, StackFrame, StoredVariableValue

__all__ = ['install_dump_frames_on', 'Traceback', 'GenerationPolicy', 'StoredVariableValue',
           'StackFrame']



import typing as tp


def one_of(*args) -> tp.Callable[[tp.Any], bool]:
    """
    Return a predicate that will return True if passed value equals to one of the arguments

    :param args: a list of arguments on which the predicate will return True
    """
    def predicate(v):
        return v in args
    return predicate



import itertools
import typing as tp
from collections import namedtuple


def parallel_for(cursor, query: tp.Union[tp.List[str], str, 'Statement', tp.List['Statement']],
                 arguments: tp.Iterable[tuple]) -> tp.Iterator[namedtuple]:
    """
    Syntactic sugar for

    >>> futures = []
    >>> for args in arguments:
    >>>     futures.append(cursor.execute_async(query, args))
    >>> for future in futures:
    >>>     yield future.result()

    If query is a string or a Cassandra Statement, or else

    >>> futures = []
    >>> for query, args in zip(query, arguments):
    >>>     futures.append(cursor.execute_async(query, args))
    >>> for future in futures:
    >>>     yield future.result()

    Note that if None is encountered in the argument iterable, session.execute() will
    be called with a single argument. You better have it as a BoundStatement then!

    :param cursor: the Cassandra cursor to use (obtained using connection.session())
    :param query: base query or a list of queries, if a different one is to be used
    :param arguments: iterable yielding arguments to use in execute_async
    """
    try:
        from cassandra.query import Statement
        query_classes = (str, Statement)
    except ImportError:
        query_classes = str

    if isinstance(query, query_classes):
        query = itertools.repeat(query)

    futures = []
    for query, args in zip(query, arguments):
        if args is None:
            future = cursor.execute_async(query)
        else:
            future = cursor.execute_async(query, args)
        futures.append(future)

    for future in futures:
        yield future.result()



import copy
import typing as tp

from satella.coding.recast_exceptions import rethrow_as
from satella.configuration.schema import Descriptor, descriptor_from_dict
from satella.exceptions import ConfigurationValidationError

T = tp.TypeVar('T')


class DictObject(tp.MutableMapping[str, T]):
    """
    A dictionary wrapper that can be accessed by attributes.

    You can use keys different than strings, but they will be inaccessible as attributes, and
    you will have to do subscription to get them.

    Eg:

    >>> a = DictObject({'test': 5})
    >>> self.assertEqual(a.test, 5)
    """

    def __contains__(self, item: str) -> bool:
        return item in self.__data

    def __str__(self) -> str:
        return 'DictObject(%s)' % (str(self.__data),)

    def __repr__(self) -> str:
        return 'DictObject(%s)' % (repr(self.__data),)

    def setdefault(self, k: str, default: tp.Any) -> tp.Any:
        try:
            return self.__data[k]
        except KeyError:
            self.__data[k] = default
            return default

    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            if isinstance(args[0], DictObject):
                args = args[0].__data,
        self.__data = dict(*args, **kwargs)

    def __delitem__(self, k: str) -> None:
        del self.__data[k]

    def __setitem__(self, k: str, v: T) -> None:
        self.__data[k] = v

    def __getitem__(self, item: str) -> T:
        return self.__data[item]

    def __iter__(self) -> tp.Iterator[str]:
        return iter(self.__data)

    def __len__(self) -> int:
        return len(self.__data)

    def __copy__(self) -> 'DictObject':
        return DictObject(self.__data.copy())

    def __eq__(self, other: dict):
        if isinstance(other, DictObject):
            return self.__data == other.__data
        else:
            return self.__data == other

    def copy(self) -> 'DictObject':
        return DictObject(self.__data.copy())

    def __deepcopy__(self, memo) -> 'DictObject':
        return DictObject(copy.deepcopy(self.__data, memo))

    @rethrow_as(KeyError, AttributeError)
    def __getattr__(self, item: str) -> T:
        return self[item]

    def __setattr__(self, key: str, value: T) -> None:
        if key == '_DictObject__data':
            return super().__setattr__(key, value)
        else:
            self[key] = value

    @rethrow_as(KeyError, AttributeError)
    def __delattr__(self, key: str) -> None:
        del self[key]

    def is_valid_schema(self, schema: tp.Optional[tp.Union[Descriptor, tp.Dict]] = None,
                        **kwarg_schema) -> bool:
        """
        Check if this dictionary conforms to particular schema.

        Schema is either a Descriptor, or a JSON-based schema. See satella.configuration.schema for details.
        Schema can be passed as well using kwargs. Note that the schema argument will be ignored if kwargs are passed.

        :param schema: schema to verify against
        :return: whether is conformant
        """

        if kwarg_schema:
            schema = kwarg_schema

        if isinstance(schema, Descriptor):
            descriptor = schema
        else:
            descriptor = descriptor_from_dict(schema)

        try:
            descriptor(self.__data)
        except ConfigurationValidationError:
            return False
        else:
            return True


def apply_dict_object(v: tp.Union[tp.Any, tp.Dict[str, T]]) -> tp.Union[DictObject, tp.Any]:
    """
    Apply DictObject() to every dict inside v.

    This assumes that the only things that will be touched will be nested dicts and lists.

    If you pass a non-dict and a non-list, they will be returned as is.
    """
    if isinstance(v, list):
        return [apply_dict_object(x) for x in v]
    elif isinstance(v, dict):
        return DictObject({
            k: apply_dict_object(val) for k, val in v.items()
        })
    else:
        return v



import collections
import time
import typing as tp
import warnings

from .base import EmbeddedSubmetrics
from .registry import register_metric
from ..data import MetricData, MetricDataCollection


@register_metric
class ClicksPerTimeUnitMetric(EmbeddedSubmetrics):
    """
    This tracks the amount of calls to handle() during the last time periods, as specified by time_unit_vectors
    (in seconds). You may specify multiple time periods as consequent entries in the list.

    By default (if you do not specify otherwise) this will track calls made during the last second.
    """
    __slots__ = ('last_clicks', 'aggregate_children', 'cutoff_period', 'time_unit_vectors')

    CLASS_NAME = 'cps'

    def __init__(self, *args, time_unit_vectors: tp.Optional[tp.List[float]] = None,
                 aggregate_children: bool = True, internal: bool = False, **kwargs):
        warnings.warn('cps is deprecated and will be removed in Satella 3.0, '
                      'use a counter and calculate a rate() from it instead',
                      DeprecationWarning)
        super().__init__(*args, internal=internal, time_unit_vectors=time_unit_vectors, **kwargs)
        time_unit_vectors = time_unit_vectors or [1]
        self.last_clicks = collections.deque()  # type: tp.List[float]
        self.aggregate_children = aggregate_children  # type: bool
        self.cutoff_period = max(time_unit_vectors)  # type: int
        self.time_unit_vectors = time_unit_vectors  # type: tp.List[int]

    def _handle(self, **labels) -> None:
        if labels or self.embedded_submetrics_enabled:
            return super()._handle(**labels)

        mono_time = time.monotonic()
        self.last_clicks.append(time.monotonic())
        try:
            while self.last_clicks[0] <= mono_time - self.cutoff_period:
                self.last_clicks.popleft()
        except IndexError:
            pass

    def to_metric_data(self) -> tp.List[int]:
        if self.embedded_submetrics_enabled:
            k = super().to_metric_data()
            if not self.aggregate_children:
                return k
            else:
                last_clicks = []
                for child in self.children:
                    last_clicks.extend(child.last_clicks)

                sum_data = self.count_vectors(last_clicks)
                sum_data.postfix_with('total')

                return k + sum_data

        return self.count_vectors(self.last_clicks)

    def count_vectors(self, last_clicks) -> MetricDataCollection:
        count_map = [0] * len(self.time_unit_vectors)
        mono_time = time.monotonic()
        time_unit_vectors = [mono_time - v for v in self.time_unit_vectors]

        for v in last_clicks:
            for index, cutoff in enumerate(time_unit_vectors):
                if v >= cutoff:
                    count_map[index] += 1

        output = []
        for time_unit, count in zip(self.time_unit_vectors, count_map):
            output.append(MetricData(self.name, count, {'period': time_unit, **self.labels},
                                     self.get_timestamp(), self.internal))

        return MetricDataCollection(output)



import heapq
import time
import typing as tp

from satella.coding.recast_exceptions import rethrow_as
from .base import Heap

Number = tp.Union[int, float]
T = tp.TypeVar('T')


class TimeBasedHeap(Heap):
    """
    A heap of items sorted by timestamps.

    It is easy to ask for items, whose timestamps are LOWER than a value, and
    easy to remove them.

    Can be used to implement a scheduling service, ie. store jobs, and each
    interval query
    which of them should be executed. This loses time resolution, but is fast.

    Can use current time with put/pop_less_than.
    Use default_clock_source to pass a callable:

    * time.time
    * time.monotonic

    Default is time.monotonic

    #notthreadsafe
    """

    def __repr__(self) -> str:
        return '<satella.coding.TimeBasedHeap with %s elements>' % (len(self.data),)

    def items(self) -> tp.Iterable[tp.Tuple[Number, T]]:
        """
        Return an iterable, but WITHOUT timestamps (only items), in
        unspecified order
        """
        return (ob for ts, ob in self.data)

    def __init__(self, default_clock_source: tp.Callable[[], Number] = time.monotonic):
        """
        Initialize an empty heap
        """
        self.default_clock_source = default_clock_source
        super().__init__(from_list=())

    def pop_timestamp(self, timestamp: Number) -> T:
        """
        Get first item with given timestamp, while maintaining the heap invariant

        :raise ValueError: element not found
        """
        for index, item in enumerate(self.data):
            if item[0] == timestamp:
                ts, item = self.data.pop(index)
                heapq.heapify(self.data)
                return item
        raise ValueError('Element not found!')

    def get_timestamp(self, item: T) -> Number:
        """
        Return the timestamp for given item
        """
        for ts, elem in self.data:
            if elem[1] == item:
                return ts

    def pop_item(self, item: T) -> tp.Tuple[Number, T]:
        """
        Pop an item off the heap, maintaining the heap invariant.

        The item will be a second part of the tuple

        :raise ValueError: element not found
        """
        for index, elem in enumerate(self.data):
            if elem[1] == item:
                obj = self.data.pop(index)
                heapq.heapify(self.data)
                return obj
        raise ValueError('Element not found!')

    def put(self, timestamp_or_value: tp.Union[T, Number],
            value: tp.Optional[T] = None) -> None:
        """
        Put an item on heap.

        Pass timestamp, item or just an item for default time
        """
        if value is None:
            timestamp, item = self.default_clock_source(), timestamp_or_value
        else:
            timestamp, item = timestamp_or_value, value

        assert timestamp is not None
        self.push((timestamp, item))

    def pop_less_than(self, less: tp.Optional[Number] = None) -> tp.Iterator[tp.Tuple[Number, T]]:
        """
        Return all elements less (sharp inequality) than particular value.

        This changes state of the heap

        :param less: value to compare against. If left at default, it will be the
            default clock source specified at construction.
        :return: an Iterator
        """

        if less is None:
            less = self.default_clock_source()

        assert less is not None, 'Default clock source returned None!'

        while self:
            if self.data[0][0] >= less:
                return
            yield self.pop()

    def remove(self, item: T) -> None:
        """
        Remove all things equal to item
        """
        self.filter_map(filter_fun=lambda i: i[1] != item)


class TimeBasedSetHeap(Heap):
    """
    A heap of items sorted by timestamps, with such invariant that every item can appear at most once.

    Note that elements you insert in this must be eq-able and hashable, ie. you can put them in a dict.

    It is easy to ask for items, whose timestamps are LOWER than a value, and
    easy to remove them.

    Can be used to implement a scheduling service, ie. store jobs, and each
    interval query
    which of them should be executed. This loses time resolution, but is fast.

    Can use current time with put/pop_less_than.
    Use default_clock_source to pass a callable:

    * time.time
    * time.monotonic

    Default is time.monotonic

    #notthreadsafe
    """

    def __repr__(self) -> str:
        return '<satella.coding.TimeBasedSetHeap with %s elements>' % (len(self.data),)

    @rethrow_as(KeyError, ValueError)
    def get_timestamp(self, item: T) -> Number:
        """
        Return the timestamp for given item

        :raises ValueError: item not found
        """
        return self.item_to_timestamp[item]

    def items(self) -> tp.Iterator[T]:
        """
        Return an iterator, but WITHOUT timestamps (only items), in
        unspecified order
        """
        return (ob for ts, ob in self.data)

    def __init__(self, default_clock_source: tp.Callable[[], Number] = None):
        """
        Initialize an empty heap
        """
        self.default_clock_source = default_clock_source or time.monotonic
        super().__init__(from_list=())
        self.item_to_timestamp = {}

    def pop_timestamp(self, timestamp: Number) -> T:
        """
        Pop an arbitary object (in case there's two) item with given timestamp,
        while maintaining the heap invariant

        :raise ValueError: element not found
        """
        for index, item in enumerate(self.data):
            if item[0] == timestamp:
                obj = self.data.pop(index)
                del self.item_to_timestamp[obj[1]]
                heapq.heapify(self.data)
                return obj[1]
        raise ValueError('Element not found!')

    def pop_item(self, item: T) -> tp.Tuple[Number, T]:
        """
        Pop an item off the heap, maintaining the heap invariant.

        The item will be a second part of the tuple

        :raise ValueError: element not found
        """
        for index, elem in enumerate(self.data):
            if elem[1] == item:
                obj = self.data.pop(index)
                heapq.heapify(self.data)
                del self.item_to_timestamp[obj[1]]
                return obj
        raise ValueError('Element not found!')

    def push(self, item: tp.Tuple[Number, T]) -> None:
        if item[1] in self.item_to_timestamp:
            self.pop_item(item[1])

        super().push(item)
        self.item_to_timestamp[item[1]] = item[0]

    def pop(self) -> tp.Tuple[Number, T]:
        item = super().pop()
        del self.item_to_timestamp[item[1]]
        return item

    def put(self, timestamp_or_value: tp.Union[T, Number],
            value: tp.Optional[T] = None) -> None:
        """
        Put an item on heap.

        Pass timestamp, item or just an item for default time
        """
        if value is None:
            timestamp, item = self.default_clock_source(), timestamp_or_value
        else:
            timestamp, item = timestamp_or_value, value

        assert timestamp is not None
        self.push((timestamp, item))

    def pop_less_than(self, less: tp.Optional[Number] = None) -> tp.Iterator[tp.Tuple[Number, T]]:
        """
        Return all elements less (sharp inequality) than particular value.

        This changes state of the heap

        :param less: value to compare against
        :return: a Generator
        """

        if less is None:
            less = self.default_clock_source()

        assert less is not None, 'Default clock source returned None!'

        while self:
            if self.data[0][0] >= less:
                return
            yield self.pop()

    def remove(self, item: T) -> None:
        """
        Remove all things equal to item
        """
        self.filter_map(filter_fun=lambda i: i[1] != item)



from .arguments import auto_adapt_to_methods, attach_arguments, for_argument, \
    execute_before
from .decorators import wraps, queue_get, chain_functions, has_keys, short_none
from .preconditions import postcondition, precondition

__all__ = ['execute_before', 'postcondition', 'precondition', 'wraps', 'queue_get', 'chain_functions',
           'has_keys', 'short_none', 'auto_adapt_to_methods', 'attach_arguments', 'for_argument']
